@@ -17,6 +17,8 @@
 package Authen::Radius;
 
 use strict;
+use warnings;
+use v5.10;
 use FileHandle;
 use IO::Socket;
 use IO::Select;
@@ -37,7 +39,7 @@ require Exporter;
             STATUS_SERVER
             COA_REQUEST COA_ACCEPT COA_REJECT COA_ACK COA_NAK);
 
-$VERSION = '0.25';
+$VERSION = '0.28';
 
 my (%dict_id, %dict_name, %dict_val, %dict_vendor_id, %dict_vendor_name );
 my ($request_id) = $$ & 0xff;   # probably better than starting from 0
@@ -375,10 +377,9 @@ sub clear_attributes {
 }
 
 sub _decode_enum {
-    my ( $name, $value, $format ) = @_;
+    my ( $name, $value) = @_;
 
-    $value = unpack( $format, $value );
-    if ( defined( $dict_val{$name}{$value} ) ) {
+    if ( defined $value && defined( $dict_val{$name}{$value} ) ) {
         $value = $dict_val{$name}{$value}{name};
     }
 
@@ -386,18 +387,48 @@ sub _decode_enum {
 }
 
 sub _decode_string {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $name, $value, $has_tag ) = @_;
 
     if ( $id == 2 && $vendor eq NO_VENDOR ) {
         return '<encrypted>';
     }
 
-    return $value;
+    if ($has_tag) {
+        my $tag = unpack('C', substr($value, 0, 1));
+        # rfc2868 section-3.3
+        # If the Tag field is greater than 0x1F, it SHOULD be
+        # interpreted as the first byte of the following String field.
+        if ($tag > 31) {
+            print STDERR "Attribute $name has tag value $tag bigger than 31 - ignoring it!\n" if $debug;
+            $tag = undef;
+        }
+        else {
+            # cut extracted tag
+            substr($value, 0, 1, '');
+        }
+        return ($value, $tag);
+    }
+
+    return ($value);
 }
 
 sub _decode_integer {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
-    return _decode_enum( $name, $value, 'N' );
+    my ( $self, $vendor, $id, $name, $value, $has_tag ) = @_;
+
+    my $tag;
+    if ($has_tag) {
+        $tag = unpack('C', substr($value, 0, 1));
+        if ($tag > 31) {
+            print STDERR "Attribute $name has tag value $tag bigger than 31 - ignoring it!\n" if $debug;
+            $tag = undef;
+        }
+        else {
+            substr($value, 0, 1, "\x00");
+        }
+    }
+
+    $value = unpack('N', $value);
+    return (_decode_enum( $name, $value), $tag);
 }
 
 sub _decode_ipaddr {
@@ -474,6 +505,11 @@ sub _decode_sublist {
     return join( '; ', @values );
 }
 
+sub _decode_octets {
+    my ( $self, $vendor, $id, $name, $value ) = @_;
+    return '0x'.unpack("H*", $value);
+}
+
 my %decoder = (
     # RFC2865
     string  => \&_decode_string,
@@ -481,6 +517,7 @@ my %decoder = (
     ipaddr  => \&_decode_ipaddr,
     date    => \&_decode_integer,
     time    => \&_decode_integer,
+    octets  => \&_decode_octets,
     # RFC3162
     ipv6addr   => \&_decode_ipv6addr,
     ipv6prefix => \&_decode_ipv6prefix,
@@ -493,11 +530,12 @@ my %decoder = (
 );
 
 sub _decode_value {
-    my ( $self, $vendor, $id, $type, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $type, $name, $value, $has_tag ) = @_;
 
     if ( defined $type ) {
         if ( exists $decoder{$type} ) {
-            return $decoder{$type}->( $self, $vendor, $id, $name, $value );
+            my ($decoded, $tag) = $decoder{$type}->( $self, $vendor, $id, $name, $value, $has_tag );
+            return wantarray ? ($decoded, $tag) : $decoded;
         }
         else {
             if ($debug) {
@@ -537,12 +575,8 @@ sub get_attributes {
 
         $name  = $r->{name} // $id;
         $type  = $r->{type};
-        $tag = undef;
-        if ($r->{has_tag}) {
-            ( $tag, $rawvalue ) = unpack('Ca*', $rawvalue);
-        }
 
-        $value = $self->_decode_value( $vendor, $id, $type, $name, $rawvalue );
+        ($value, $tag) = $self->_decode_value( $vendor, $id, $type, $name, $rawvalue, $r->{has_tag} );
 
         push(
             @a, {
@@ -584,7 +618,7 @@ sub _encode_enum {
 }
 
 sub _encode_string {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $name, $value, $tag ) = @_;
 
     if ( $id == 2 && $vendor eq NO_VENDOR ) {
         $self->gen_authenticator();
@@ -597,12 +631,21 @@ sub _encode_string {
     #   return pack('C', 0) . substr($_[0], 0, 246);
     # }
 
+    if (defined $tag) {
+        $value = pack('C', $tag) . $value;
+    }
+
     return $value;
 }
 
 sub _encode_integer {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
-    return _encode_enum( $name, $value, 'N' );
+    my ( $self, $vendor, $id, $name, $value, $tag ) = @_;
+    $value = _encode_enum( $name, $value, 'N' );
+    if (defined $tag) {
+        # tag added to 1st byte, not extending the value length
+        substr($value, 0, 1, pack('C', $tag) );
+    }
+    return $value;
 }
 
 sub _encode_ipaddr {
@@ -690,7 +733,7 @@ sub _encode_sublist {
     $value = '';
     foreach my $da (@pairs) {
         my ( $subname, $subvalue ) = @{$da};
-        my $subid = $dict_val{$id}->{$subname}->{id};
+        my $subid = $dict_val{$name}->{$subname}->{id};
         next if ( !defined($subid) );
         $value .= pack( 'CC', $subid, length($subvalue) + 2 ) . $subvalue;
     }
@@ -782,11 +825,11 @@ my %encoder = (
 );
 
 sub _encode_value {
-    my ( $self, $vendor, $id, $type, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $type, $name, $value, $tag ) = @_;
 
     if ( defined $type ) {
         if ( exists $encoder{$type} ) {
-            return $encoder{$type}->( $self, $vendor, $id, $name, $value );
+            return $encoder{$type}->( $self, $vendor, $id, $name, $value, $tag );
         }
         else {
             if ($debug) {
@@ -820,6 +863,8 @@ sub add_attributes {
             $attr_name = $1;
         }
 
+        die 'unknown attr name '.$attr_name if (! exists $dict_name{$attr_name});
+
         $id = $dict_name{$attr_name}{id} // int($attr_name);
         $vendor = vendorID($attr);
         if (exists($dict_name{$attr_name}{'tlv'})) {
@@ -864,28 +909,29 @@ sub add_attributes {
             # WiMAX uses non-standard VSAs - include the continuation byte
         }
 
-        unless (defined($value = $self->_encode_value($vendor, $id, $type, $a->{Name}, $a->{Value}))) {
+        unless (defined($value = $self->_encode_value($vendor, $id, $type, $a->{Name}, $a->{Value}, $a->{Tag}))) {
             print STDERR "Unable to encode attribute $a->{Name} ($id, $type, $vendor) with value '$a->{Value}'\n" if $debug;
             next;
         }
-        print STDERR "Adding attribute $a->{Name} ($id, $type, $vendor) with value '$a->{Value}'\n" if $debug;
+
+        if ($debug) {
+            printf STDERR "Adding attribute %s (%s, %s, %s) with value '%s'%s\n",
+                    $a->{Name}, $id, $type, $vendor,
+                    $a->{Value},
+                    ($a->{Tag} ? sprintf(' (tag:%d)', $a->{Tag}) : '');
+        }
+
         if ( $vendor eq NO_VENDOR ) {
-            if ($need_tag) {
-                $self->{'attributes'} .= pack('C C C', $id, length($value) + 3, $a->{Tag} // 0) . $value;
-            }
-            else {
-                $self->{'attributes'} .= pack('C C', $id, length($value) + 2) . $value;
-            }
+            # tag already included in $value, if any
+            $self->{'attributes'} .= pack('C C', $id, length($value) + 2) . $value;
         } else {
             # VSA
             # pack vendor-ID + vendor-type + vendor-length
             if ($vendor eq WIMAX_VENDOR) {
                 # add continuation byte
                 $value = pack('N C C C', $vendor, $id, length($value) + 3, 0) . $value;
-            } elsif ($need_tag) {
-                # tagged attribute
-                $value = pack('N C C C', $vendor, $id, length($value) + 3, $a->{Tag} // 0) . $value;
             } else {
+                # tag already included in $value, if any
                 $value = pack('N C C', $vendor, $id, length($value) + 2) . $value;
             }
 
@@ -1044,6 +1090,11 @@ sub load_dictionary {
             $dict_vendor_id{$id}{'name'} = $name;
         } elsif ($cmd eq 'begin-vendor') {
             $dict_def_vendor = $name;
+            if (! $freeradius_dict) {
+                # force format
+                $freeradius_dict = 1;
+                print STDERR "Detected BEGIN-VENDOR, switch to FreeRADIUS dictionary format\n" if $debug;
+            }
         } elsif ($cmd eq 'end-vendor') {
             $dict_def_vendor = NO_VENDOR;
         } elsif ($cmd eq 'begin-tlv') {
@@ -1058,7 +1109,7 @@ sub load_dictionary {
             my @path = split("/", $file);
             pop @path; # remove the filename at the end
             my $path = ( $name =~ /^\// ) ? $name : join("/", @path, $name);
-            load_dictionary('', $path);
+            load_dictionary('', $path, %opt);
         }
     }
     $fh->close;
